/********************************************************************************
*                                                                               *
*             D y n a m i c   L i n k   L i b r a r y   S u p p o r t           *
*                                                                               *
*********************************************************************************
* Copyright (C) 2002,2006 by Jeroen van der Zijp.   All Rights Reserved.        *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.    *
*********************************************************************************
* $Id: FXDLL.cpp,v 1.35 2006/07/19 21:50:07 fox Exp $                           *
********************************************************************************/
#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "FXHash.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXDLL.h"
#ifndef WIN32
#ifdef HAVE_SHL_LOAD
#include <dl.h>                 // HP-UX
#else
#include <dlfcn.h>              // POSIX
#endif
#endif
#ifndef RTLD_GLOBAL
#define RTLD_GLOBAL 0           // Does not exist on DEC
#endif
#ifdef HAVE_SHL_LOAD
#ifndef	DYNAMIC_PATH            // HP-UX
#define DYNAMIC_PATH 0
#endif
#ifndef	BIND_RESTRICTED
#define BIND_RESTRICTED	0
#endif
#endif

/*
  Notes:
  - Make sure it works on other unices.
*/

using namespace FX;

/*******************************************************************************/

namespace FX {


// Return name of the DLL
FXString FXDLL::name() const {
  if(hnd){
#ifdef WIN32
    char modulename[1024];
    if(GetModuleFileNameA((HINSTANCE)hnd,modulename,sizeof(modulename))){
      return modulename;
      }
#else
#ifdef HAVE_SHL_LOAD    // HP-UX
    // FIXME
#else                   // POSIX
    Dl_info info;
    if(dladdr(hnd,&info)){
      //FXTRACE((1,"dli_fname = %s\n",info.dli_fname));
      //FXTRACE((1,"dli_fbase = %p\n",info.dli_fbase));
      //FXTRACE((1,"dli_sname = %s\n",info.dli_sname));
      //FXTRACE((1,"dli_saddr = %p\n",info.dli_saddr));
      return info.dli_fname;
      }
#endif
#endif
    }
  return FXString::null;
  }


// Load the dll given name
FXbool FXDLL::load(const FXString& nam){
  if(!hnd){
#ifndef WIN32
#ifdef HAVE_SHL_LOAD    // HP-UX
    hnd=shl_load(nam.text(),BIND_IMMEDIATE|BIND_NONFATAL|DYNAMIC_PATH,0L);
#else
#ifdef DL_LAZY		// OpenBSD
    hnd=dlopen(nam.text(),DL_LAZY);
#else			// POSIX
    hnd=dlopen(nam.text(),RTLD_NOW|RTLD_GLOBAL);
#endif
#endif
#else                   // WIN32
    // Order of loading with LoadLibrary (or LoadLibraryEx with no
    // LOAD_WITH_ALTERED_SEARCH_PATH flag):
    //
    // 1. Directory from which the application was loaded.
    // 2. Current directory.
    // 3. System directory, as determined by GetSystemDirectory().
    // 4. 16-bit system directory.
    // 5. Windows directory, as determined by GetWindowsDirectory().
    // 6. Directories in the $PATH.
    //
    // With flag LOAD_WITH_ALTERED_SEARCH_PATH:
    //
    // 1. Directory specified by the filename path.
    // 2. Current directory.
    // 3. System directory.
    // 4. 16-bit system directory.
    // 5. Windows directory.
    // 6. Directories in the $PATH.
    //
    // We switched to the latter so sub-modules needed by a DLL are
    // plucked from the same place as name (thanks to Rafael de
    // Pelegrini Soares" <Rafael@enq.ufrgs.br>).
    hnd=LoadLibraryExA(nam.text(),NULL,LOAD_WITH_ALTERED_SEARCH_PATH);
#endif
    }
  return hnd!=NULL;
  }


// Unload the dll
void FXDLL::unload(){
  if(hnd){
#ifndef WIN32
#ifdef HAVE_SHL_LOAD    // HP-UX
    shl_unload((shl_t)hnd);
#else			// POSIX
    dlclose(hnd);
#endif
#else                   // WIN32
    FreeLibrary((HMODULE)hnd);
#endif
    hnd=NULL;
    }
  }


// Resolve a symbol
void* FXDLL::symbol(const FXchar* nam) const {
  if(hnd){
#ifndef WIN32
#ifdef HAVE_SHL_LOAD    // HP-UX
    void* address=NULL;
    if(shl_findsym((shl_t*)&hnd,nam,TYPE_UNDEFINED,&address)==0) return address;
#else			// POSIX
    return dlsym(hnd,nam);
#endif
#else                   // WIN32
    return (void*)GetProcAddress((HMODULE)hnd,nam);
#endif
    }
  return NULL;
  }


// Resolve a symbol
void* FXDLL::symbol(const FXString& nam) const {
  return symbol(nam.text());
  }


// Find DLL of ourselves
FXDLL FXDLL::dll(){
  return dll((void*)FXDLL::error);
  }


// Find DLL containing symbol
FXDLL FXDLL::dll(void* symbol){
#ifdef WIN32
  MEMORY_BASIC_INFORMATION mbi;
  if(VirtualQuery((const void*)symbol,&mbi,sizeof(mbi))){
    //FXTRACE((1,"BaseAddress       = %p\n",mbi.BaseAddress));
    //FXTRACE((1,"AllocationBase    = %p\n",mbi.AllocationBase));
    //FXTRACE((1,"AllocationProtect = 0x%x\n",mbi.AllocationProtect));
    //FXTRACE((1,"RegionSize        = %d\n",mbi.RegionSize));
    //FXTRACE((1,"State             = 0x%x\n",mbi.State));
    //FXTRACE((1,"Protect           = 0x%x\n",mbi.Protect));
    //FXTRACE((1,"Type              = 0x%x\n",mbi.Type));
    return FXDLL(mbi.AllocationBase);
    }
#else
#ifdef HAVE_SHL_LOAD    // HP-UX
  // FIXME //
#else                   // POSIX
  Dl_info info;
  if(dladdr(symbol,&info)){
    //FXTRACE((1,"dli_fname = %s\n",info.dli_fname));
    //FXTRACE((1,"dli_fbase = %p\n",info.dli_fbase));
    //FXTRACE((1,"dli_sname = %s\n",info.dli_sname));
    //FXTRACE((1,"dli_saddr = %p\n",info.dli_saddr));
    return FXDLL(info.dli_fbase);
    }
#endif
#endif
  return FXDLL(NULL);
  }


// Return error message if error occurred loading a DLL
FXString FXDLL::error(){
#ifndef WIN32
#ifdef HAVE_SHL_LOAD    // HP-UX
  return FXString::null;
#else			// POSIX
  return dlerror();
#endif
#else                   // WIN32
  DWORD dw=GetLastError();
  FXchar buffer[512];
  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,NULL,dw,MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),(LPTSTR)buffer,sizeof(buffer),NULL);
  return buffer;
#endif
  }


// Initialize by loading given library name
FXAUTODLL::FXAUTODLL(const FXString& nam){
  load(nam);
  }


// Unload library if we have one
FXAUTODLL::~FXAUTODLL(){
  unload();
  }

}

